@@ -36,6 +36,7 @@ if TYPE_CHECKING:
     from sqlalchemy.engine.base import Connection
     from sqlalchemy.engine.base import Transaction
     from sqlalchemy.engine.mock import MockConnection
+    from sqlalchemy.sql.elements import ClauseElement
 
     from .environment import EnvironmentContext
     from ..config import Config
@@ -49,11 +50,11 @@ log = logging.getLogger(__name__)
 
 
 class _ProxyTransaction:
-    def __init__(self, migration_context: "MigrationContext") -> None:
+    def __init__(self, migration_context: MigrationContext) -> None:
         self.migration_context = migration_context
 
     @property
-    def _proxied_transaction(self) -> Optional["Transaction"]:
+    def _proxied_transaction(self) -> Optional[Transaction]:
         return self.migration_context._transaction
 
     def rollback(self) -> None:
@@ -68,7 +69,7 @@ class _ProxyTransaction:
         t.commit()
         self.migration_context._transaction = None
 
-    def __enter__(self) -> "_ProxyTransaction":
+    def __enter__(self) -> _ProxyTransaction:
         return self
 
     def __exit__(self, type_: None, value: None, traceback: None) -> None:
@@ -126,22 +127,22 @@ class MigrationContext:
 
     def __init__(
         self,
-        dialect: "Dialect",
-        connection: Optional["Connection"],
+        dialect: Dialect,
+        connection: Optional[Connection],
         opts: Dict[str, Any],
-        environment_context: Optional["EnvironmentContext"] = None,
+        environment_context: Optional[EnvironmentContext] = None,
     ) -> None:
         self.environment_context = environment_context
         self.opts = opts
         self.dialect = dialect
-        self.script: Optional["ScriptDirectory"] = opts.get("script")
+        self.script: Optional[ScriptDirectory] = opts.get("script")
         as_sql: bool = opts.get("as_sql", False)
         transactional_ddl = opts.get("transactional_ddl")
         self._transaction_per_migration = opts.get(
             "transaction_per_migration", False
         )
         self.on_version_apply_callbacks = opts.get("on_version_apply", ())
-        self._transaction: Optional["Transaction"] = None
+        self._transaction: Optional[Transaction] = None
 
         if as_sql:
             self.connection = cast(
@@ -214,14 +215,14 @@ class MigrationContext:
     @classmethod
     def configure(
         cls,
-        connection: Optional["Connection"] = None,
+        connection: Optional[Connection] = None,
         url: Optional[str] = None,
         dialect_name: Optional[str] = None,
-        dialect: Optional["Dialect"] = None,
-        environment_context: Optional["EnvironmentContext"] = None,
+        dialect: Optional[Dialect] = None,
+        environment_context: Optional[EnvironmentContext] = None,
         dialect_opts: Optional[Dict[str, str]] = None,
         opts: Optional[Any] = None,
-    ) -> "MigrationContext":
+    ) -> MigrationContext:
         """Create a new :class:`.MigrationContext`.
 
         This is a factory method usually called
@@ -365,7 +366,7 @@ class MigrationContext:
 
     def begin_transaction(
         self, _per_migration: bool = False
-    ) -> Union["_ProxyTransaction", ContextManager]:
+    ) -> Union[_ProxyTransaction, ContextManager]:
         """Begin a logical transaction for migration operations.
 
         This method is used within an ``env.py`` script to demarcate where
@@ -539,6 +540,7 @@ class MigrationContext:
 
     def _ensure_version_table(self, purge: bool = False) -> None:
         with sqla_compat._ensure_scope_for_ddl(self.connection):
+            assert self.connection is not None
             self._version.create(self.connection, checkfirst=True)
             if purge:
                 assert self.connection is not None
@@ -550,9 +552,7 @@ class MigrationContext:
             self.connection, self.version_table, self.version_table_schema
         )
 
-    def stamp(
-        self, script_directory: "ScriptDirectory", revision: str
-    ) -> None:
+    def stamp(self, script_directory: ScriptDirectory, revision: str) -> None:
         """Stamp the version table with a specific revision.
 
         This method calculates those branches to which the given revision
@@ -568,7 +568,7 @@ class MigrationContext:
         for step in script_directory._stamp_revs(revision, heads):
             head_maintainer.update_to_step(step)
 
-    def run_migrations(self, **kw) -> None:
+    def run_migrations(self, **kw: Any) -> None:
         r"""Run the migration scripts established for this
         :class:`.MigrationContext`, if any.
 
@@ -614,6 +614,7 @@ class MigrationContext:
                 if self.as_sql and not head_maintainer.heads:
                     # for offline mode, include a CREATE TABLE from
                     # the base
+                    assert self.connection is not None
                     self._version.create(self.connection)
                 log.info("Running %s", step)
                 if self.as_sql:
@@ -637,6 +638,7 @@ class MigrationContext:
                     )
 
         if self.as_sql and not head_maintainer.heads:
+            assert self.connection is not None
             self._version.drop(self.connection)
 
     def _in_connection_transaction(self) -> bool:
@@ -647,7 +649,11 @@ class MigrationContext:
         else:
             return meth()
 
-    def execute(self, sql: str, execution_options: None = None) -> None:
+    def execute(
+        self,
+        sql: Union[ClauseElement, str],
+        execution_options: Optional[dict] = None,
+    ) -> None:
         """Execute a SQL construct or string statement.
 
         The underlying execution mechanics are used, that is
@@ -659,15 +665,15 @@ class MigrationContext:
         self.impl._exec(sql, execution_options)
 
     def _stdout_connection(
-        self, connection: Optional["Connection"]
-    ) -> "MockConnection":
+        self, connection: Optional[Connection]
+    ) -> MockConnection:
         def dump(construct, *multiparams, **params):
             self.impl._exec(construct)
 
         return MockEngineStrategy.MockConnection(self.dialect, dump)
 
     @property
-    def bind(self) -> Optional["Connection"]:
+    def bind(self) -> Optional[Connection]:
         """Return the current "bind".
 
         In online mode, this is an instance of
@@ -688,7 +694,7 @@ class MigrationContext:
         return self.connection
 
     @property
-    def config(self) -> Optional["Config"]:
+    def config(self) -> Optional[Config]:
         """Return the :class:`.Config` used by the current environment,
         if any."""
 
@@ -698,7 +704,7 @@ class MigrationContext:
             return None
 
     def _compare_type(
-        self, inspector_column: "Column", metadata_column: "Column"
+        self, inspector_column: Column, metadata_column: Column
     ) -> bool:
         if self._user_compare_type is False:
             return False
@@ -718,8 +724,8 @@ class MigrationContext:
 
     def _compare_server_default(
         self,
-        inspector_column: "Column",
-        metadata_column: "Column",
+        inspector_column: Column,
+        metadata_column: Column,
         rendered_metadata_default: Optional[str],
         rendered_column_default: Optional[str],
     ) -> bool:
@@ -748,7 +754,7 @@ class MigrationContext:
 
 
 class HeadMaintainer:
-    def __init__(self, context: "MigrationContext", heads: Any) -> None:
+    def __init__(self, context: MigrationContext, heads: Any) -> None:
         self.context = context
         self.heads = set(heads)
 
@@ -771,9 +777,11 @@ class HeadMaintainer:
                 == literal_column("'%s'" % version)
             )
         )
+
         if (
             not self.context.as_sql
             and self.context.dialect.supports_sane_rowcount
+            and ret is not None
             and ret.rowcount != 1
         ):
             raise util.CommandError(
@@ -796,9 +804,11 @@ class HeadMaintainer:
                 == literal_column("'%s'" % from_)
             )
         )
+
         if (
             not self.context.as_sql
             and self.context.dialect.supports_sane_rowcount
+            and ret is not None
             and ret.rowcount != 1
         ):
             raise util.CommandError(
@@ -808,7 +818,7 @@ class HeadMaintainer:
                 % (from_, to_, self.context.version_table, ret.rowcount)
             )
 
-    def update_to_step(self, step: Union["RevisionStep", "StampStep"]) -> None:
+    def update_to_step(self, step: Union[RevisionStep, StampStep]) -> None:
         if step.should_delete_branch(self.heads):
             vers = step.delete_version_num
             log.debug("branch delete %s", vers)
@@ -904,12 +914,12 @@ class MigrationInfo:
     from dependencies.
     """
 
-    revision_map: "RevisionMap"
+    revision_map: RevisionMap
     """The revision map inside of which this operation occurs."""
 
     def __init__(
         self,
-        revision_map: "RevisionMap",
+        revision_map: RevisionMap,
         is_upgrade: bool,
         is_stamp: bool,
         up_revisions: Union[str, Tuple[str, ...]],
@@ -998,14 +1008,14 @@ class MigrationStep:
 
     @classmethod
     def upgrade_from_script(
-        cls, revision_map: "RevisionMap", script: "Script"
-    ) -> "RevisionStep":
+        cls, revision_map: RevisionMap, script: Script
+    ) -> RevisionStep:
         return RevisionStep(revision_map, script, True)
 
     @classmethod
     def downgrade_from_script(
-        cls, revision_map: "RevisionMap", script: "Script"
-    ) -> "RevisionStep":
+        cls, revision_map: RevisionMap, script: Script
+    ) -> RevisionStep:
         return RevisionStep(revision_map, script, False)
 
     @property
@@ -1034,7 +1044,7 @@ class MigrationStep:
 
 class RevisionStep(MigrationStep):
     def __init__(
-        self, revision_map: "RevisionMap", revision: "Script", is_upgrade: bool
+        self, revision_map: RevisionMap, revision: Script, is_upgrade: bool
     ) -> None:
         self.revision_map = revision_map
         self.revision = revision
@@ -1130,12 +1140,12 @@ class RevisionStep(MigrationStep):
         other_heads = set(heads).difference(self.from_revisions)
 
         if other_heads:
-            ancestors = set(
+            ancestors = {
                 r.revision
                 for r in self.revision_map._get_ancestor_nodes(
                     self.revision_map.get_revisions(other_heads), check=False
                 )
-            )
+            }
             from_revisions = list(
                 set(self.from_revisions).difference(ancestors)
             )
@@ -1152,12 +1162,12 @@ class RevisionStep(MigrationStep):
     def _unmerge_to_revisions(self, heads: Collection[str]) -> Tuple[str, ...]:
         other_heads = set(heads).difference([self.revision.revision])
         if other_heads:
-            ancestors = set(
+            ancestors = {
                 r.revision
                 for r in self.revision_map._get_ancestor_nodes(
                     self.revision_map.get_revisions(other_heads), check=False
                 )
-            )
+            }
             return tuple(set(self.to_revisions).difference(ancestors))
         else:
             return self.to_revisions
@@ -1241,7 +1251,7 @@ class RevisionStep(MigrationStep):
         return self.revision.revision
 
     @property
-    def info(self) -> "MigrationInfo":
+    def info(self) -> MigrationInfo:
         return MigrationInfo(
             revision_map=self.revision_map,
             up_revisions=self.revision.revision,
@@ -1258,7 +1268,7 @@ class StampStep(MigrationStep):
         to_: Optional[Union[str, Collection[str]]],
         is_upgrade: bool,
         branch_move: bool,
-        revision_map: Optional["RevisionMap"] = None,
+        revision_map: Optional[RevisionMap] = None,
     ) -> None:
         self.from_: Tuple[str, ...] = util.to_tuple(from_, default=())
         self.to_: Tuple[str, ...] = util.to_tuple(to_, default=())
@@ -1269,7 +1279,7 @@ class StampStep(MigrationStep):
 
     doc: None = None
 
-    def stamp_revision(self, **kw) -> None:
+    def stamp_revision(self, **kw: Any) -> None:
         return None
 
     def __eq__(self, other):
@@ -1356,7 +1366,7 @@ class StampStep(MigrationStep):
         return len(self.to_) > 1
 
     @property
-    def info(self) -> "MigrationInfo":
+    def info(self) -> MigrationInfo:
         up, down = (
             (self.to_, self.from_)
             if self.is_upgrade
